﻿Option Strict On
Option Infer On

Imports Helpers.Mathematics
Imports Helpers.Direct2D
Imports SlimDX
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports System.IO
Imports System.Reflection

Friend MustInherit Class Creep
    Inherits Entity

    Private life As Integer
    Private maxLife As Integer
    Private col As Color
    Private colors() As Color
    Private m_speed As Double
    Private way As WaySegment
    Private points() As PointF
    Private pIndex As Integer
    Private sIndex As Integer
    Private slowedDown As Boolean
    Private slowdownTime As Integer
    Private slowDownMaxTime As Integer
    Private slowDownFactor As Double
    Private shieldValue As Integer
    Private speedFactor As Double
    Private dodgeChance As Integer
    Private blockedTower As Type

    Friend Shared creepINI As INIFile

    Friend Shared Sub Init()
        creepINI = New INIFile(Path.Combine(Game.getEnvironmentPath("settings"), "creeps.ini"))
        Dim creeps = GetAllCreeps()
        For Each c In creeps
            CreepRenderer.RegisterRenderer(c.ID, Sub(e, rect) c.CreepType.GetMethod("Render").Invoke(Nothing, {e, rect}))
        Next
    End Sub

    Friend Shared Function GetAllCreeps() As CreepDescriptor()
        Dim currentAssembly = Assembly.GetExecutingAssembly
        Dim currentTypes = currentAssembly.GetTypes
        Dim creeps As New List(Of Type)
        Dim genericCreepDefinitions As New List(Of Type)
        For Each t In currentTypes
            Dim constructors() = t.GetConstructors
            Dim hasRightConstructor = True
            For Each c In constructors
                If c.GetParameters.Count = 0 Then
                    hasRightConstructor = True
                    Exit For
                End If
            Next
            If hasRightConstructor AndAlso t.IsSubclassOf(GetType(Creep)) Then
                If t.IsGenericTypeDefinition Then
                    genericCreepDefinitions.Add(t)
                Else
                    creeps.Add(t)
                End If
            End If
        Next
        Dim descriptors As New List(Of CreepDescriptor)
        For i = 0 To creeps.Count - 1
            descriptors.Add(New CreepDescriptor(creeps(i), CShort(i), 0))
        Next
        For i = 0 To genericCreepDefinitions.Count - 1
            For j = 0 To creeps.Count - 1
                descriptors.Add(New CreepDescriptor(genericCreepDefinitions(i).MakeGenericType(creeps(j)), CShort(j), CShort(i + 1)))
            Next
        Next
        Return descriptors.ToArray()
    End Function

    Friend Sub BlockTower(t As Type)
        blockedTower = t
    End Sub

    Friend Sub UnblockTower()
        blockedTower = Nothing
    End Sub

    Protected ReadOnly Property angle As Single
        Get
            If way Is Nothing Then Return 0
            Return Convert.ToSingle(MathHelper.getAngle(New Helpers.Mathematics.Vector2(way.P1), New Helpers.Mathematics.Vector2(way.P2)))
        End Get
    End Property

    Friend ReadOnly Property speed As Double
        Get
            Return m_speed
        End Get
    End Property

    Friend Sub AktivateShield()
        shieldValue = Convert.ToInt32(maxLife * 0.2)
    End Sub

    Friend Sub DeaktivateShield()
        shieldValue = 0
    End Sub

    Friend Sub ApllySpeedModifier(factor As Double)
        speedFactor = factor
    End Sub

    Friend Sub RemoveSpeedModifier()
        speedFactor = 1
    End Sub

    Friend Sub SetDodgeChance(percent As Integer)
        dodgeChance = percent
    End Sub

    Friend Sub ResetDodgeChance()
        dodgeChance = 0
    End Sub

    Friend Sub SlowDown(factor As Double, time As Integer)
        slowedDown = True
        slowdownTime = 0
        slowDownMaxTime = time
        slowDownFactor = factor
        points = way.GetPoints(m_speed * speedFactor * factor, getLocation, way.P2)
        pIndex = 0
    End Sub

    Friend Overrides Sub doStep()
        If slowedDown Then
            slowdownTime += 1
            If slowdownTime = slowDownMaxTime Then
                slowedDown = False
                slowdownTime = 0
                slowDownFactor = 1
                points = way.GetPoints(m_speed * speedFactor, getLocation, way.P2)
                pIndex = 0
            End If
        End If
        If way IsNot Nothing Then pIndex += 1
        If points Is Nothing Then Return
        If pIndex >= points.Length - 1 Then
            sIndex += 1
            way.onCreepReachEnd(Me)
            way = Nothing
            If world IsNot Nothing AndAlso sIndex < world.way.segments.Length Then
                way = world.way.segments(sIndex)
                way.onCreepReachStart(Me)
                points = way.GetPoints(m_speed * speedFactor * slowDownFactor)
            Else
                Return
            End If
            pIndex = 0
        End If
    End Sub

    Friend Shared ReadOnly Property size As SizeF
        Get
            Return New SizeF(22, 22)
        End Get
    End Property

    Friend Overrides Function getSize() As SizeF
        Return Creep.size
    End Function

    Friend Overrides Function getLocation() As PointF
        If points Is Nothing OrElse pIndex > points.Length - 1 Then Return Nothing
        Return points(pIndex)
    End Function

    Friend Function getLife() As Integer
        Return life
    End Function

    Friend Sub MoveTo(c As Creep)
        Do While InUse
            Threading.Thread.Sleep(1)
        Loop
        InUse = True
        sIndex = c.sIndex
        way.onCreepReachEnd(Me)
        way = world.way.segments(sIndex)
        way.onCreepReachStart(Me)
        points = way.GetPoints(m_speed * speedFactor * slowDownFactor, c.getLocation, way.P2)
        pIndex = 0
        InUse = False
    End Sub

    Sub New(life As Integer, colors() As Color, c As Color, s As Double)
        Me.life = life
        maxLife = life
        Me.colors = colors
        col = c
        m_speed = s
        speedFactor = 1
        slowDownFactor = 1
    End Sub

    Friend Overrides Sub onSpawned()
        MyBase.onSpawned()
        way = World.way.segments(0)
        pIndex = 0
        sIndex = 0
        life += Convert.ToInt32(life * World.currentWave * 0.1)
        maxLife = life
        way.onCreepReachStart(Me)
        points = way.GetPoints(m_speed * speedFactor)
    End Sub

    Friend Sub takeDamage(damage As Integer, towerType As Type)
        If towerType IsNot Nothing AndAlso towerType Is blockedTower Then Return
        If dodgeChance > 0 AndAlso Game.rand.Next(100) < dodgeChance Then
            Dim ind As New DamageIndicator("Ausgewichen", getLocation)
            world.spawnEntity(ind)
            Return
        End If
        If shieldValue >= damage Then
            shieldValue -= damage
        Else
            life -= (damage - shieldValue)
            shieldValue = 0
        End If
        'Dim di As New DamageIndicator(damage, getLocation)
        'world.spawnEntity(di)
        If life <= 0 Then destroy()
    End Sub

    Protected Sub destroy()
        World.addMoney(2)
        Dim e As New Explosion(getLocation, colors, col, 80, 25.0F, True, True)
        World.spawnEntity(e)
        World.despawnEntity(Me)
    End Sub

    Friend Function containsPoint(p As PointF) As Boolean
        Dim loc As PointF = getLocation()
        Dim rect As New RectangleF(loc.X - size.Width / 2.0F, loc.Y - size.Height / 2.0F, size.Width, size.Height)
        Dim pRect As New RectangleF(p.X, p.Y, 1, 1)
        Return rect.IntersectsWith(pRect)
    End Function

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        If shieldValue > 0 Then
            Dim shieldBrushName = "ShieldBrush"
            If Not ColorManager.BrushRegistered(shieldBrushName) Then ColorManager.RegisterBrush(shieldBrushName, New SolidColorBrush(e.Target, New Color4(Color.SkyBlue)))
            Dim shieldEllipse As New Ellipse
            shieldEllipse.Center = New PointF(offset.X + getLocation.X, offset.Y + getLocation.Y)
            shieldEllipse.RadiusX = size.Width * 0.75F
            shieldEllipse.RadiusY = size.Height * 0.75F
            ColorManager.GetBrush(shieldBrushName).Opacity = 0.3F
            e.Target.FillEllipse(ColorManager.GetBrush(shieldBrushName), shieldEllipse)
            ColorManager.GetBrush(shieldBrushName).Opacity = 1
            e.Target.DrawEllipse(ColorManager.GetBrush(shieldBrushName), shieldEllipse, 2)
        End If
        Dim percent = Math.Max(Convert.ToSingle(life / maxLife), 0)
        Dim g = percent
        Dim r = 1.0F - percent
        Dim c As New Color4(1, r, g, 0)
        Dim brushName = "LifeBarBrush" & c.ToString
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, c))
        Dim rect As New RectangleF(getLocation.X - size.Width / 2.0F + offset.X - 5, getLocation.Y - size.Height / 2.0F + offset.Y - 14.0F, size.Width + 10, 3)
        e.Target.FillRectangle(ColorManager.GetBrush("defaultBlack"), rect)
        e.Target.FillRectangle(ColorManager.GetBrush(brushName), New RectangleF(rect.Location, New SizeF(rect.Width * percent, 3)))
        If shieldValue > 0 Then
            Dim shieldPercent = Convert.ToSingle(shieldValue / Convert.ToInt32(maxLife * 0.2))
            Dim shieldBarBrushName = "ShieldBarBrush"
            If Not ColorManager.BrushRegistered(shieldBarBrushName) Then ColorManager.RegisterBrush(shieldBarBrushName, New SolidColorBrush(e.Target, New Color4(Color.LightBlue)))
            Dim shieldRect As New RectangleF(getLocation.X - size.Width / 2.0F + offset.X - 5, getLocation.Y - size.Height / 2.0F + offset.Y - 17.0F, size.Width + 10, 3)
            e.Target.FillRectangle(ColorManager.GetBrush("defaultBlack"), shieldRect)
            e.Target.FillRectangle(ColorManager.GetBrush(shieldBarBrushName), New RectangleF(shieldRect.Location, New SizeF(shieldRect.Width * shieldPercent, 3)))
        End If
    End Sub
End Class

Friend Class CreepDescriptor
    Private _creepType As Type
    Private _id As Integer

    Friend ReadOnly Property CreepType As Type
        Get
            Return _creepType
        End Get
    End Property

    Friend ReadOnly Property ID As Integer
        Get
            Return _id
        End Get
    End Property

    Friend Sub New(creepType As Type, baseID As Short, subID As Short)
        _creepType = creepType
        _id = (CInt(baseID) << 16) Or subID
    End Sub
End Class

Friend Class StarCreep
    Inherits Creep

    Private starAngle As Integer

    Sub New()
        MyBase.New(Integer.Parse(creepINI.getValue("c01_life")), {Color.Red, Color.LightBlue, Color.Lime, Color.Pink, Color.Yellow, Color.Orange, Color.Blue, Color.Purple}, Color.Pink, Double.Parse(creepINI.getValue("c01_speed"), System.Globalization.CultureInfo.InvariantCulture.NumberFormat))
        starAngle = 0
    End Sub

    Friend Overrides Sub doStep()
        MyBase.doStep()
        starAngle += 3
        If starAngle >= 360 Then starAngle = starAngle Mod 360
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        MyBase.draw(e, offset)
        DrawInner(e, offset, getLocation, size.Width / 2.0F, starAngle)
    End Sub

    Public Shared Sub Render(e As D2DControlset.D2DDrawEventArgs, rect As RectangleF)
        DrawInner(e, New PointF(0, 0), New PointF(rect.X + rect.Width / 2.0F, rect.Y + rect.Height / 2.0F), Math.Min(rect.Width, rect.Height) / 2.0F, 90)
    End Sub

    Private Shared Sub DrawInner(e As D2DControlset.D2DDrawEventArgs, offset As PointF, loc As PointF, radius As Single, angle As Single)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim brushName = "StarCreepBrush"
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New Color4(Color.Pink)))
        h.DrawStar(New PointF(loc.X + offset.X, loc.Y + offset.Y), radius, angle, ColorManager.GetBrush(brushName), 2, 5)
    End Sub
End Class

Friend Class SquareCreep
    Inherits Creep

    Sub New()
        MyBase.New(Integer.Parse(creepINI.getValue("c02_life")), {Color.Red, Color.LightBlue, Color.Lime, Color.Pink, Color.Yellow, Color.Orange, Color.Blue, Color.Purple}, Color.Red, Double.Parse(creepINI.getValue("c02_speed"), System.Globalization.CultureInfo.InvariantCulture.NumberFormat))
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        MyBase.draw(e, offset)
        DrawInner(e, offset, getLocation, size.Width / 2.0F, angle)
    End Sub

    Public Shared Sub Render(e As D2DControlset.D2DDrawEventArgs, rect As RectangleF)
        DrawInner(e, New PointF(0, 0), New PointF(rect.X + rect.Width / 2.0F, rect.Y + rect.Height / 2.0F), Math.Min(rect.Width, rect.Height) / 2.0F, 0)
    End Sub

    Private Shared Sub DrawInner(e As D2DControlset.D2DDrawEventArgs, offset As PointF, loc As PointF, radius As Single, angle As Single)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim brushName = "SquareCreepBrush"
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New Color4(Color.Red)))
        h.drawSqare(New PointF(loc.X + offset.X, loc.Y + offset.Y), radius, Convert.ToSingle(MathHelper.getDeg(angle)), ColorManager.GetBrush(brushName), 2)
    End Sub
End Class

Friend Class PacmanCreep
    Inherits Creep

    Sub New()
        MyBase.New(Integer.Parse(creepINI.getValue("c03_life")), {Color.Red, Color.LightBlue, Color.Lime, Color.Pink, Color.Yellow, Color.Orange, Color.Blue, Color.Purple}, Color.LightBlue, Double.Parse(creepINI.getValue("c03_speed"), System.Globalization.CultureInfo.InvariantCulture.NumberFormat))
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        MyBase.draw(e, offset)
        DrawInner(e, offset, getLocation, size.Width / 2.0F, angle)
    End Sub

    Public Shared Sub Render(e As D2DControlset.D2DDrawEventArgs, rect As RectangleF)
        DrawInner(e, New PointF(0, 0), New PointF(rect.X + rect.Width / 2.0F, rect.Y + rect.Height / 2.0F), Math.Min(rect.Width, rect.Height) / 2.0F, Math.PI)
    End Sub

    Private Shared Sub DrawInner(e As D2DControlset.D2DDrawEventArgs, offset As PointF, loc As PointF, radius As Single, angle As Single)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim brushName = "PacmanCreepBrush"
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New Color4(Color.LightBlue)))
        h.DrawPie(New PointF(loc.X + offset.X, loc.Y + offset.Y), radius, Convert.ToSingle(MathHelper.getDeg(angle)) + 200, 320, ColorManager.GetBrush(brushName), 2)
    End Sub
End Class

Friend Class EllipseCreep
    Inherits Creep

    Sub New()
        MyBase.New(Integer.Parse(creepINI.getValue("c04_life")), {Color.Red, Color.LightBlue, Color.Lime, Color.Pink, Color.Yellow, Color.Orange, Color.Blue, Color.Purple}, Color.Orange, Double.Parse(creepINI.getValue("c04_speed"), System.Globalization.CultureInfo.InvariantCulture.NumberFormat))
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        MyBase.draw(e, offset)
        DrawInner(e, offset, getLocation, size.Width / 2.0F, angle)
    End Sub

    Public Shared Sub Render(e As D2DControlset.D2DDrawEventArgs, rect As RectangleF)
        DrawInner(e, New PointF(0, 0), New PointF(rect.X + rect.Width / 2.0F, rect.Y + rect.Height / 2.0F), Math.Min(rect.Width, rect.Height) / 2.0F, 0)
    End Sub

    Private Shared Sub DrawInner(e As D2DControlset.D2DDrawEventArgs, offset As PointF, loc As PointF, radius As Single, angle As Single)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim brushName = "EllipseCreepBrush"
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New Color4(Color.OrangeRed)))
        Dim el As New Ellipse
        el.Center = New PointF(loc.X + offset.X, loc.Y + offset.Y)
        el.RadiusX = radius
        el.RadiusY = radius * 0.7F
        h.DrawEllipse(el, Convert.ToSingle(MathHelper.getDeg(angle)), ColorManager.GetBrush(brushName), 2)
    End Sub
End Class

Friend Class CircleCreep
    Inherits Creep

    Sub New()
        MyBase.New(Integer.Parse(creepINI.getValue("c05_life")), {Color.Red, Color.LightBlue, Color.Lime, Color.Pink, Color.Yellow, Color.Orange, Color.Blue, Color.Purple}, Color.Yellow, Double.Parse(creepINI.getValue("c05_speed"), System.Globalization.CultureInfo.InvariantCulture.NumberFormat))
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        MyBase.draw(e, offset)
        DrawInner(e, offset, getLocation, size.Width / 2.0F)
    End Sub

    Public Shared Sub Render(e As D2DControlset.D2DDrawEventArgs, rect As RectangleF)
        DrawInner(e, New PointF(0, 0), New PointF(rect.X + rect.Width / 2.0F, rect.Y + rect.Height / 2.0F), Math.Min(rect.Width, rect.Height) / 2.0F)
    End Sub

    Private Shared Sub DrawInner(e As D2DControlset.D2DDrawEventArgs, offset As PointF, loc As PointF, radius As Single)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim brushName = "CircleCreepBrush"
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New Color4(Color.Yellow)))
        h.DrawCircle(New Helpers.Mathematics.Vector2(loc.X + offset.X, loc.Y + offset.Y), radius, ColorManager.GetBrush(brushName), 2)
    End Sub
End Class

Friend Class BossCreep(Of T As {New, Creep}) : Inherits Creep
    Private R, G, B As Byte
    Private changeMode As Integer
    Private angleDeg As Integer
    Private triangleSize As Integer
    Private sizeDir As Boolean
    Private steps As Integer

    Sub New()
        MyBase.New(Integer.Parse(creepINI.getValue("c06_life")), {Color.Red, Color.LightBlue, Color.Lime, Color.Pink, Color.Yellow, Color.Orange, Color.Blue, Color.Purple}, Nothing, Double.Parse(creepINI.getValue("c06_speed"), System.Globalization.CultureInfo.InvariantCulture.NumberFormat))
        R = Byte.MaxValue
        sizeDir = True
        triangleSize = 60
    End Sub

    Friend Overrides Sub doStep()
        MyBase.doStep()
        Select Case changeMode
            Case 0
                R -= CByte(1)
                G += CByte(1)
                If G = Byte.MaxValue Then changeMode = 1
            Case 1
                G -= CByte(1)
                B += CByte(1)
                If B = Byte.MaxValue Then changeMode = 2
            Case 2
                B -= CByte(1)
                R += CByte(1)
                If R = Byte.MaxValue Then changeMode = 0
            Case Else
                R = Byte.MaxValue
                G = 0
                B = 0
                changeMode = 0
        End Select
        angleDeg += 1
        If angleDeg = 360 Then angleDeg = 0
        If sizeDir Then
            triangleSize += 1
            If triangleSize = 100 Then sizeDir = False
        Else
            triangleSize -= 1
            If triangleSize = 60 Then sizeDir = True
        End If
        steps += 1
        If steps Mod 240 = 0 Then
            Dim c As New T
            world.spawnEntity(c)
            c.MoveTo(Me)
        End If
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        MyBase.draw(e, offset)
        DrawInner(e, offset, getLocation, triangleSize / 4.0F, (100 - (triangleSize - 60)) / 4.0F, angleDeg, R, G, B)
    End Sub

    Public Shared Sub Render(e As D2DControlset.D2DDrawEventArgs, rect As RectangleF)
        DrawInner(e, New PointF(0, 0), New PointF(rect.X + rect.Width / 2.0F, rect.Y + rect.Height / 2.0F), Math.Min(rect.Width, rect.Height) / 2.0F, Math.Min(rect.Width, rect.Height) / 2.5F, 90, 128, 0, 128)
    End Sub

    Private Shared Sub DrawInner(e As D2DControlset.D2DDrawEventArgs, offset As PointF, loc As PointF, radius1 As Single, radius2 As Single, angle As Single, r As Byte, g As Byte, b As Byte)
        Dim brushName = String.Join(",", "bossCreepColor", r, g, b)
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New Color4(Color.FromArgb(r, g, b))))
        Dim triangle1(2) As PointF
        Dim triangle2(2) As PointF
        Dim m As New Helpers.Mathematics.Vector2(loc.X + offset.X, loc.Y + offset.Y)
        For i = 0 To 2
            Dim deg1 = angle + i * 120
            Dim deg2 = deg1 + 60
            triangle1(i) = MathHelper.getPointOnCircle(m, MathHelper.getRad(deg1), radius1).toPointF
            triangle2(i) = MathHelper.getPointOnCircle(m, MathHelper.getRad(deg2), radius2).toPointF
        Next
        Dim pg As New PathGeometry(e.Factory)
        Dim sink = pg.Open
        sink.BeginFigure(triangle1.First, FigureBegin.Hollow)
        sink.AddLines(triangle1.Skip(1).ToArray)
        sink.EndFigure(FigureEnd.Closed)
        sink.BeginFigure(triangle2.First, FigureBegin.Hollow)
        sink.AddLines(triangle2.Skip(1).ToArray)
        sink.EndFigure(FigureEnd.Closed)
        sink.Close()
        e.Target.DrawGeometry(pg, ColorManager.GetBrush(brushName), 2)
        sink.Dispose()
        pg.Dispose()
    End Sub
End Class

Friend Class CreepRenderer
    Private Shared RendererList As Dictionary(Of Integer, Action(Of D2DControlset.D2DDrawEventArgs, RectangleF))

    Shared Sub New()
        RendererList = New Dictionary(Of Integer, Action(Of D2DControlset.D2DDrawEventArgs, RectangleF))
    End Sub

    Friend Shared Sub RegisterRenderer(id As Integer, renderer As Action(Of D2DControlset.D2DDrawEventArgs, RectangleF))
        RendererList.Add(id, renderer)
    End Sub

    Friend Shared Sub DrawCreep(id As Integer, e As D2DControlset.D2DDrawEventArgs, rect As RectangleF)
        RendererList(id).Invoke(e, rect)
    End Sub
End Class

Friend Class DamageIndicator
    Inherits Entity

    Private value As String
    Private location As PointF
    Private steps As Integer
    Private height As Single
    Private opacity As Single

    Friend Overrides Sub doStep()
        steps += 1
        height = steps * 0.5F
        If steps > 90 Then
            opacity = 1 - ((steps - 90) / 30.0F)
        End If
        If steps > 120 Then world.despawnEntity(Me)
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        Dim b As New SolidColorBrush(e.Target, New Color4(1.0F, 0, 0))
        b.Opacity = opacity
        Dim format = e.TextFactory.CreateTextFormat("Segoe UI", FontWeight.SemiBold, FontStyle.Oblique, FontStretch.Normal, 16.0F, Game.locale)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim s = h.MeasureString(value, format, New SizeF(Single.MaxValue, Single.MaxValue))
        s.Width += 1
        Dim rect As New RectangleF(location.X - s.Width / 2.0F + offset.X, location.Y - s.Height / 2.0F + offset.Y - height, s.Width, s.Height)
        e.Target.DrawText(value, format, rect, b)
        format.Dispose()
        b.Dispose()
    End Sub

    Sub New(val As String, loc As PointF)
        value = val
        location = loc
        steps = 0
        height = 0.0F
        opacity = 1
    End Sub

    Friend Overrides Function getLocation() As PointF
        Return location
    End Function

    Friend Overrides Function getSize() As SizeF
        Return New Size(0, 0)
    End Function
End Class