﻿Option Strict On
Option Infer On

Imports Helpers.Mathematics
Imports SlimDX
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports D2DControlset
Imports System.IO
Imports System.Runtime.InteropServices
Imports System.Text
Imports System.Threading
Imports MessageBoxExtended
Imports System.Runtime.CompilerServices

Module Game
    Friend WithEvents window As GameWindow
    Private environmentPaths As Dictionary(Of String, String)
    Private TDLogo As Bitmap
    Private visibleMenu As Menu
    Friend lastMenuChange As DateTime
    Private r As Random
    Private worlds As List(Of World)
    Private nonUserMapBMP As Bitmap
    Friend towerImages() As Bitmap
    Friend toolImages() As Bitmap
    Private m_globalVolume As UShort
    Friend guiINI As INIFile
    Friend handlingBMP As Bitmap

    'Hauptmenü Steuerelemente
    Private WithEvents mainMenu As Menu
    Private mmPic As D2DPictureBox
    Private versionLabel As D2DLabel
    Private WithEvents mmButton1 As TDButton
    Private WithEvents mmButton2 As TDButton
    Private WithEvents mmButton3 As TDButton
    Private WithEvents mmButton4 As TDButton
    Private WithEvents mmButton5 As TDButton
    Private WithEvents mmButton6 As TDButton

    'Einstellungen Steuerelemente
    Private WithEvents optionsMenu As Menu
    Private WithEvents globalVolumeBar As TDTrackBar
    Private WithEvents musicVolumeBar As TDTrackBar
    Private WithEvents effectVolumeBar As TDTrackBar
    Private WithEvents optionsButton As TDButton
    Private omLabel1 As D2DLabel
    Private omLabel2 As D2DLabel
    Private omLabel3 As D2DLabel
    Private WithEvents fullscreenButton As TDButton
    Private omHeader As D2DLabel

    'Einzelspieler Steuerelemente
    Private WithEvents singleplayerMenu As Menu
    Private WithEvents mapList As TDListBox
    Private spmHeader As D2DLabel
    Private WithEvents spmButton1 As TDButton
    Private WithEvents spmButton2 As TDButton
    Private smpDescTextBox As TDTextBox

    'Welteditor Steuerelemente
    Private WithEvents worldEditMenu As Menu
    Private WithEvents weMapList As TDListBox
    Private weHeader As D2DLabel
    Private WithEvents weButton1 As TDButton
    Private WithEvents weButton2 As TDButton

    Private cControl As CampaignControl

    Private updateButton As IngameButton

    <DllImport("winmm.dll", SetLastError:=True, CallingConvention:=CallingConvention.Winapi)> _
    Private Function waveOutSetVolume(uDeviceID As IntPtr, dwVolume As UInteger) As Integer
    End Function

    Private Sub setVolume(volume As UShort)
        Dim bytes As New List(Of Byte)(BitConverter.GetBytes(volume))
        bytes.AddRange(bytes)
        Dim uInt = BitConverter.ToUInt32(bytes.ToArray, 0)
        waveOutSetVolume(IntPtr.Zero, uInt)
    End Sub

    Private ReadOnly Property target As RenderTarget
        Get
            Return window.getTarget
        End Get
    End Property

    Private ReadOnly Property factory As Direct2D.Factory
        Get
            Return window.getFactory
        End Get
    End Property

    Private ReadOnly Property textFactory As DirectWrite.Factory
        Get
            Return window.getTextFactory
        End Get
    End Property

    Friend ReadOnly Property rand As Random
        Get
            Return r
        End Get
    End Property

    Friend Property globalVolume As UShort
        Get
            Return m_globalVolume
        End Get
        Set(value As UShort)
            m_globalVolume = value
            setVolume(m_globalVolume)
        End Set
    End Property

    Friend ReadOnly Property locale As String
        Get
            Return guiINI.getValue("locale", "en-us")
        End Get
    End Property

    <Extension()>
    Public Function Offset(first As PointF, second As PointF) As PointF
        Return New PointF(first.X + second.X, first.Y + second.Y)
    End Function

    <STAThread>
    Sub main()

        'Mutex erstellen
        Dim firstInstance As Boolean
        Dim m As New Mutex(False, "TowerDefenseSingleInstanceMutex", firstInstance)
        If Not firstInstance Then
            m.Close()
            m = Nothing
            Exit Sub
        End If

        Application.EnableVisualStyles()

        'Fehlerhandhabung
        AddHandler Application.ThreadException, AddressOf onError
        Application.SetUnhandledExceptionMode(UnhandledExceptionMode.CatchException)
        
        'Zufallsgenerator initialisieren
        r = New Random

        'Umgebungspfade festlegen
        environmentPaths = New Dictionary(Of String, String)
        Dim dir = Environment.CurrentDirectory
        environmentPaths.Add("img", Path.Combine(dir, "resources\img"))
        environmentPaths.Add("cur", Path.Combine(dir, "resources\cur"))
        environmentPaths.Add("musicconf", Path.Combine(dir, "config\music"))
        environmentPaths.Add("sounds", Path.Combine(dir, "resources\sounds"))
        environmentPaths.Add("settings", Path.Combine(dir, "config\settings"))
        environmentPaths.Add("maps", Path.Combine(dir, "maps"))
        environmentPaths.Add("tutorialmaps", Path.Combine(dir, "tutorial"))
        environmentPaths.Add("campaignmaps", Path.Combine(dir, "campaign"))

        'GUI-Texte laden
        guiINI = New INIFile(Path.Combine(getEnvironmentPath("settings"), "gui.ini"))

        'Fenster vorbereiten
        window = New GameWindow
        window.Text = "TowerDefense2"
        window.CanSize = True
        window.MaximizeBox = True
        window.Size = New Size(1296, 758)
        window.StartPosition = FormStartPosition.CenterScreen
        window.CloseBox = False
        window.Icon = My.Resources.TD
        window.ChangeCursor(0)

        'Audio vorbereiten
        MusicManager.init()
        SoundEffect.init()

        'Welteditor vorbereiten
        WorldEditControl.init()

        'Spieldaten einlesen
        Tower.init()
        Creep.Init()

        'Welten laden
        reloadWorlds()

        'Anwendung starten
        Application.Run(window)

        'Klassen entladen
        SoundEffect.unloadAll()
        ColorManager.ClearAll()
        FontManager.ClearAll()

    End Sub

    ' ----- wenn ein Fehler in der Anwendung auftritt ----- '
    Private Sub onError(sender As Object, e As ThreadExceptionEventArgs)
        showError(e.Exception)
        shutDown()
    End Sub

    ' ----- wenn die Anwendung geladen wird ----- '
    Private Sub onGameLoad(sender As Object, e As EventArgs) Handles window.HandleCreated

        'Textrendering einstellen
        target.TextAntialiasMode = TextAntialiasMode.ClearType

        'Fensterstil erstellen
        Dim lColor As New Color4(1, 0, 0, 0)
        Dim uColor As New Color4(1, 0.2F, 0.2F, 0.2F)
        Dim uiColor As New Color4(1, 0.9F, 0.9F, 0.9F)
        Dim style As New Direct2DWindow.D2DWindowStyle(window, uiColor, New Color4(1, 0, 0, 0), {uColor, uColor, uColor, uColor}, {lColor, lColor, lColor, lColor}, uiColor)
        window.SetWindowStyle(style, False)

        'Brushes erstellen
        ColorManager.RegisterBrush("defaultWhite", New SolidColorBrush(target, New Color4(1, 0.9F, 0.9F, 0.9F)))
        ColorManager.RegisterBrush("defaultBlack", New SolidColorBrush(target, New Color4(1, 0, 0, 0)))

        'Fonts erstellen
        FontManager.RegisterFont("defaultFormat", New FontTemplate("Polentical Neon", 22))
        FontManager.RegisterFont("versionFormat", New FontTemplate("Segoe UI", 16, FontStyle.Italic))
        FontManager.RegisterFont("defaultLabelFormat", New FontTemplate("Polentical Neon", 20, TextAlignment:=TextAlignment.Center, paragraphAlignment:=ParagraphAlignment.Center))
        FontManager.RegisterFont("headerFormat", New FontTemplate("Blacksmith Delight", 100, TextAlignment:=TextAlignment.Center, paragraphAlignment:=ParagraphAlignment.Center))

        Dim imgPath = getEnvironmentPath("img")
        TDLogo = loadBitmap(Path.Combine(imgPath, "Header.png"))
        nonUserMapBMP = loadBitmap(Path.Combine(imgPath, "td.png"))
        handlingBMP = loadBitmap(Path.Combine(imgPath, "Handling.png"))
        towerImages = New Bitmap(6) {}
        For i = 0 To towerImages.Length - 1
            Dim bmpPath = Path.Combine(imgPath, "t" & (i + 1).ToString("00") & ".png")
            If File.Exists(bmpPath) Then towerImages(i) = loadBitmap(bmpPath)
        Next
        toolImages = New Bitmap(2) {}
        toolImages(0) = loadBitmap(Path.Combine(imgPath, "Pen.png"))
        toolImages(1) = loadBitmap(Path.Combine(imgPath, "Eraser.png"))
        toolImages(2) = loadBitmap(Path.Combine(imgPath, "Questionmark.png"))

        Try
            Dim vals() = loadSettings()
            globalVolume = Convert.ToUInt16(vals(0))
            MusicManager.volume = vals(1)
            SoundEffect.globalVolume = vals(2)
            window.setFullscreen(Convert.ToBoolean(vals(3)))
        Catch ex As FileNotFoundException
            globalVolume = UShort.MaxValue
            MusicManager.volume = 100000
            SoundEffect.globalVolume = 100000
        End Try
        MusicManager.start()
        MusicManager.environment = MusicEnvironment.getEnvironment("menu")

        createMenu()
        changeMenu(mainMenu)
    End Sub

    ' ----- wenn die Anwendung beendet wird ----- '
    Private Sub onGameUnload(sender As Object, e As EventArgs) Handles window.Closing
        saveSettings({globalVolumeBar.Value, MusicManager.volume, SoundEffect.globalVolume, Convert.ToInt32(window.getFullscreen)})
        MusicManager.stop()
        For i = 0 To towerImages.Length - 1
            If towerImages(i) IsNot Nothing AndAlso Not towerImages(i).Disposed Then towerImages(i).Dispose()
        Next
        For i = 0 To toolImages.Length - 1
            toolImages(i).Dispose()
        Next

        window.SetWindowDefaultStyle(True)

        handlingBMP.Dispose()
        nonUserMapBMP.Dispose()
        TDLogo.Dispose()
    End Sub

    Friend Sub reloadWorlds()
        worlds = loadWorlds()
    End Sub

    Private Function loadWorlds() As List(Of World)
        Dim worlds As New List(Of World)
        Dim dir As New DirectoryInfo(getEnvironmentPath("maps"))
        For Each d As DirectoryInfo In dir.GetDirectories
            Dim w As World = Nothing
            Try
                w = New World(d.Name, MapType.SinglePlayer)
            Catch : End Try
            If w IsNot Nothing Then worlds.Add(w)
        Next
        Return worlds
    End Function

    Private Sub saveSettings(vals() As Integer)
        Dim f As String = Path.Combine(getEnvironmentPath("settings"), "settings.ini")
        If File.Exists(f) Then File.Delete(f)
        Dim fs As FileStream = File.Create(f)
        Dim bw As New BinaryWriter(fs)
        For i = 0 To vals.Length - 1
            bw.Write(vals(i))
        Next
        bw.Close()
        bw.Dispose()
        fs.Close()
        fs.Dispose()
    End Sub

    Private Function loadSettings() As Integer()
        Dim f As String = Path.Combine(getEnvironmentPath("settings"), "settings.ini")
        If Not File.Exists(f) Then Throw New FileNotFoundException("Die Datei " & f & " konnte nicht gefunden werden.")
        Dim fs As FileStream = File.Open(f, FileMode.Open, FileAccess.ReadWrite)
        Dim br As New BinaryReader(fs)
        Dim vals As New List(Of Integer)
        While Not fs.Position >= fs.Length
            vals.Add(br.ReadInt32)
        End While
        br.Close()
        br.Dispose()
        fs.Close()
        fs.Dispose()
        Return vals.ToArray
    End Function

    Friend Sub startWorld(ByVal w As WorldControl)
        changeMenu(w)
        Dim t As New Thread(New ThreadStart(AddressOf w.start))
        t.Name = "main game thread"
        t.IsBackground = True
        t.Start()
    End Sub

    Friend Sub showError(ByVal ex As Exception)
        MessageBoxEx.showException("Tower Defense: kritischer Spielfehler!", ex, True)
    End Sub

    Friend Function getWidthFactor() As Single
        Dim windowWidth As Integer = window.Width
        If window.ShowWindowBorder Then windowWidth -= 16
        Return windowWidth / 1280.0F
    End Function

    Friend Function getHeightFactor() As Single
        Dim windowHeight As Integer = window.Height
        If window.ShowWindowBorder Then
            windowHeight -= 30
            If Not window.WindowState = FormWindowState.Maximized Then windowHeight -= 8
        End If
        Return windowHeight / 720.0F
    End Function

    Friend Function getSizeFactor() As Single
        Return Math.Min(getWidthFactor, getHeightFactor)
    End Function

    Private Sub createMenu()

        'Hauptmenü
        mainMenu = New Menu
        mmPic = New D2DPictureBox(mainMenu)
        mmPic.SizeMode = PictureBoxSizeMode.Zoom
        mmPic.Image = TDLogo

        versionLabel = New D2DLabel(mainMenu)
        versionLabel.Text = "Alpha 2.0 Final" '"v" & My.Application.Info.Version.ToString
        ColorManager.RegisterBrush("versionLabelBrush", New SolidColorBrush(target, New Color4(Color.FromArgb(255, 208, 128, 229))))
        versionLabel.TextColor = CType(ColorManager.GetBrush("versionLabelBrush"), SolidColorBrush)
        versionLabel.Font = FontManager.GetFont("versionFormat")

        mmButton1 = New TDButton
        mmButton1.D2DParent = mainMenu
        mmButton1.Text = guiINI.getValue("singleplayer")
        AddHandler mmButton1.Click, Sub() changeMenu(singleplayerMenu)

        mmButton6 = New TDButton
        mmButton6.D2DParent = mainMenu
        mmButton6.Text = guiINI.getValue("load")

        mmButton2 = New TDButton
        mmButton2.D2DParent = mainMenu
        mmButton2.Text = guiINI.getValue("campagne")

        mmButton3 = New TDButton
        mmButton3.D2DParent = mainMenu
        mmButton3.Text = guiINI.getValue("worldeditor")
        AddHandler mmButton3.Click, Sub() changeMenu(worldEditMenu)

        mmButton4 = New TDButton
        mmButton4.D2DParent = mainMenu
        mmButton4.Text = guiINI.getValue("options")
        AddHandler mmButton4.Click, Sub() changeMenu(optionsMenu)

        mmButton5 = New TDButton
        mmButton5.D2DParent = mainMenu
        mmButton5.Text = guiINI.getValue("close")
        AddHandler mmButton5.Click, Sub() shutDown()

        updateButton = New IngameButton
        updateButton.D2DParent = mainMenu
        updateButton.Text = guiINI.getValue("update_notavailable")

        resizeMainMenu()
        AddHandler mainMenu.Resize, Sub() resizeMainMenu()

        'Einstellungen
        optionsMenu = New Menu
        optionsMenu.parentMenu = mainMenu
        globalVolumeBar = New TDTrackBar
        globalVolumeBar.D2DParent = optionsMenu
        globalVolumeBar.MaxValue = UShort.MaxValue
        globalVolumeBar.Value = globalVolume
        AddHandler globalVolumeBar.ValueChanged, _
            Sub()
                globalVolume = Convert.ToUInt16(globalVolumeBar.Value)
                If omLabel1 IsNot Nothing Then omLabel1.Text = guiINI.getValue("globalVolume").Replace("%volume%", (globalVolumeBar.Value * 100 \ UShort.MaxValue).ToString)
            End Sub
        musicVolumeBar = New TDTrackBar
        musicVolumeBar.D2DParent = optionsMenu
        musicVolumeBar.MinValue = 1
        musicVolumeBar.MaxValue = 100000
        musicVolumeBar.Value = MusicManager.volume
        AddHandler musicVolumeBar.ValueChanged, _
            Sub()
                MusicManager.volume = musicVolumeBar.Value
                If omLabel2 IsNot Nothing Then omLabel2.Text = guiINI.getValue("musicVolume").Replace("%volume%", (musicVolumeBar.Value \ 1000).ToString)
            End Sub
        effectVolumeBar = New TDTrackBar
        effectVolumeBar.D2DParent = optionsMenu
        effectVolumeBar.MinValue = 1
        effectVolumeBar.MaxValue = 100000
        effectVolumeBar.Value = SoundEffect.globalVolume
        AddHandler effectVolumeBar.ValueChanged, _
            Sub()
                SoundEffect.globalVolume = effectVolumeBar.Value
                If omLabel3 IsNot Nothing Then omLabel3.Text = guiINI.getValue("effectVolume").Replace("%volume%", (effectVolumeBar.Value \ 1000).ToString)
            End Sub
        optionsButton = New TDButton
        optionsButton.D2DParent = optionsMenu
        optionsButton.Text = guiINI.getValue("back")
        AddHandler optionsButton.Click, Sub() optionsMenu.goUp()
        omLabel1 = New D2DLabel(optionsMenu)
        omLabel1.Font = FontManager.GetFont("defaultLabelFormat")
        omLabel1.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
        omLabel1.Text = guiINI.getValue("globalVolume").Replace("%volume%", (globalVolumeBar.Value * 100 \ UShort.MaxValue).ToString)
        omLabel2 = New D2DLabel(optionsMenu)
        omLabel2.Font = FontManager.GetFont("defaultLabelFormat")
        omLabel2.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
        omLabel2.Text = guiINI.getValue("musicVolume").Replace("%volume%", (musicVolumeBar.Value \ 1000).ToString)
        omLabel3 = New D2DLabel(optionsMenu)
        omLabel3.Font = FontManager.GetFont("defaultLabelFormat")
        omLabel3.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
        omLabel3.Text = guiINI.getValue("effectVolume").Replace("%volume%", (effectVolumeBar.Value \ 1000).ToString)
        fullscreenButton = New TDButton
        fullscreenButton.D2DParent = optionsMenu
        fullscreenButton.Text = If(window.getFullscreen, guiINI.getValue("wm_Windowed"), guiINI.getValue("wm_Fullscreen"))
        omHeader = New D2DLabel(optionsMenu)
        omHeader.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
        omHeader.Text = guiINI.getValue("options")
        omHeader.Font = FontManager.GetFont("headerFormat")
        resizeOptionsMenu()
        AddHandler optionsMenu.Resize, Sub() resizeOptionsMenu()

        'Einzelspieler
        singleplayerMenu = New Menu
        singleplayerMenu.parentMenu = mainMenu

        mapList = New TDListBox
        mapList.D2DParent = singleplayerMenu
        mapList.ItemHeight = 32
        refreshMapList()

        spmHeader = New D2DLabel(singleplayerMenu)
        spmHeader.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
        spmHeader.Text = guiINI.getValue("singleplayer")
        spmHeader.Font = FontManager.GetFont("headerFormat")

        spmButton1 = New TDButton
        spmButton1.D2DParent = singleplayerMenu
        spmButton1.Text = guiINI.getValue("startGame")

        spmButton2 = New TDButton
        spmButton2.D2DParent = singleplayerMenu
        spmButton2.Text = guiINI.getValue("back")
        AddHandler spmButton2.Click, Sub() singleplayerMenu.goUp()

        smpDescTextBox = New TDTextBox(singleplayerMenu)
        smpDescTextBox.IsReadOnly = True

        resizeSingleplayerMenu()
        AddHandler singleplayerMenu.Resize, Sub() resizeSingleplayerMenu()

        'Welteditor
        worldEditMenu = New Menu
        worldEditMenu.parentMenu = mainMenu

        weHeader = New D2DLabel(worldEditMenu)
        weHeader.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
        weHeader.Text = guiINI.getValue("worldeditor")
        weHeader.Font = FontManager.GetFont("headerFormat")

        weButton1 = New TDButton
        weButton1.D2DParent = worldEditMenu
        weButton1.Text = guiINI.getValue("editWE")
        AddHandler weButton1.Click, Sub() If weMapList.SelectedIndex >= 0 AndAlso (DateTime.Now - lastMenuChange).Ticks >= 100000 Then editWorld(weMapList.SelectedItem.Text)

        weButton2 = New TDButton
        weButton2.D2DParent = worldEditMenu
        weButton2.Text = guiINI.getValue("back")
        AddHandler weButton2.Click, Sub() worldEditMenu.goUp()

        weMapList = New TDListBox
        weMapList.D2DParent = worldEditMenu
        weMapList.ItemHeight = 32
        refreshEditorList()

        resizeWorldEditMenu()
        AddHandler worldEditMenu.Resize, Sub() resizeWorldEditMenu()

        cControl = New CampaignControl
        cControl.parentMenu = mainMenu

    End Sub

    Friend Sub refreshMapList()
        mapList.Items.Clear()
        For i = 0 To worlds.Count - 1
            Dim item As New TDImageListBoxItem
            item.Text = worlds(i).name
            If Not worlds(i).isCustom Then item.Image = nonUserMapBMP
            mapList.Items.Add(item)
        Next
    End Sub

    Friend Sub refreshEditorList()
        weMapList.Items.Clear()
        weMapList.Items.Add(New TDTextOnlyListBoxItem() With {.Text = guiINI.getValue("entry_NEW")})
        For i = 0 To worlds.Count - 1
            If WorldEditControl.devMode OrElse worlds(i).isCustom Then
                Dim item As New TDTextOnlyListBoxItem
                item.Text = worlds(i).name
                weMapList.Items.Add(item)
            End If
        Next
    End Sub

    Private Sub resizeMainMenu()
        Dim pBorder As Integer = Convert.ToInt32(30 * getWidthFactor())
        mmPic.Location = New Point(pBorder, Convert.ToInt32(50 * getHeightFactor()))
        mmPic.Size = New Size(mainMenu.Width - 2 * pBorder, Convert.ToInt32(150 * getHeightFactor()))
        Dim lBorder As Integer = Convert.ToInt32(8 * getSizeFactor())
        versionLabel.Size = New Size(Convert.ToInt32(300 * getSizeFactor()), Convert.ToInt32(24 * getSizeFactor()))
        versionLabel.Location = New Point(lBorder, mainMenu.Height - versionLabel.Height)
        Dim bWidth As Integer = Convert.ToInt32(600 * getWidthFactor())
        Dim bHeight As Integer = Convert.ToInt32(50 * getHeightFactor())
        Dim bX As Integer = (mainMenu.Width - bWidth) \ 2
        mmButton1.Size = New Size(bWidth, bHeight)
        mmButton1.Location = New Point(bX, Convert.ToInt32(230 * getHeightFactor()))
        mmButton6.Size = New Size(bWidth, bHeight)
        mmButton6.Location = New Point(bX, Convert.ToInt32(295 * getHeightFactor()))
        mmButton2.Size = New Size(bWidth, bHeight)
        mmButton2.Location = New Point(bX, Convert.ToInt32(360 * getHeightFactor()))
        mmButton3.Size = New Size(bWidth, bHeight)
        mmButton3.Location = New Point(bX, Convert.ToInt32(425 * getHeightFactor()))
        mmButton4.Size = New Size(bWidth, bHeight)
        mmButton4.Location = New Point(bX, Convert.ToInt32(490 * getHeightFactor()))
        mmButton5.Size = New Size(bWidth, bHeight)
        mmButton5.Location = New Point(bX, Convert.ToInt32(600 * getHeightFactor()))
        updateButton.Size = New Size(Convert.ToInt32(250 * Game.getWidthFactor), Convert.ToInt32(35 * Game.getHeightFactor))
        Dim updateButtonBorder = Convert.ToInt32(10 * Game.getSizeFactor)
        updateButton.Location = New Point(mainMenu.Width - updateButton.Width - updateButtonBorder, mainMenu.Height - updateButton.Height - updateButtonBorder)
    End Sub

    Private Sub resizeOptionsMenu()
        Dim cWidth As Integer = Convert.ToInt32(600 * getWidthFactor())
        Dim cX As Integer = (mainMenu.Width - cWidth) \ 2
        optionsButton.Size = New Size(cWidth, Convert.ToInt32(50 * getHeightFactor()))
        optionsButton.Location = New Point(cX, Convert.ToInt32(600 * getHeightFactor()))
        Dim tHeight As Integer = Convert.ToInt32(25 * getHeightFactor())
        Dim lHeight As Integer = Convert.ToInt32(40 * getHeightFactor())
        globalVolumeBar.Size = New Size(cWidth, tHeight)
        globalVolumeBar.Location = New Point(cX, Convert.ToInt32(240 * getHeightFactor()))
        musicVolumeBar.Size = New Size(cWidth, tHeight)
        musicVolumeBar.Location = New Point(cX, Convert.ToInt32(320 * getHeightFactor()))
        effectVolumeBar.Size = New Size(cWidth, tHeight)
        effectVolumeBar.Location = New Point(cX, Convert.ToInt32(400 * getHeightFactor()))
        omLabel1.Location = New Point(cX, Convert.ToInt32(200 * getHeightFactor()))
        omLabel1.Size = New Size(cWidth, lHeight)
        omLabel2.Location = New Point(cX, Convert.ToInt32(280 * getHeightFactor()))
        omLabel2.Size = New Size(cWidth, lHeight)
        omLabel3.Location = New Point(cX, Convert.ToInt32(360 * getHeightFactor()))
        omLabel3.Size = New Size(cWidth, lHeight)
        fullscreenButton.Size = New Size(cWidth, Convert.ToInt32(50 * getHeightFactor()))
        fullscreenButton.Location = New Point(cX, Convert.ToInt32(460 * getHeightFactor()))
        omHeader.Size = New Size(optionsMenu.Width, Convert.ToInt32(optionsMenu.Height / 4.0F))
    End Sub

    Private Sub resizeSingleplayerMenu()
        spmHeader.Size = New Size(singleplayerMenu.Width, Convert.ToInt32(singleplayerMenu.Height / 4.0F))
        Dim cWidth As Integer = Convert.ToInt32(480 * getWidthFactor())
        Dim cBorder As Integer = Convert.ToInt32(150 * getWidthFactor())
        mapList.Location = New Point(cBorder, Convert.ToInt32(200 * getHeightFactor()))
        mapList.Size = New Size(cWidth, Convert.ToInt32(400 * getHeightFactor()))
        Dim bHeight As Integer = Convert.ToInt32(50 * getHeightFactor())
        Dim bX As Integer = singleplayerMenu.Width - cBorder - cWidth
        spmButton2.Size = New Size(cWidth, bHeight)
        spmButton2.Location = New Point(bX, mapList.Y + mapList.Height - bHeight)
        spmButton1.Size = New Size(cWidth, bHeight)
        spmButton1.Location = New Point(bX, spmButton2.Y - bHeight - Convert.ToInt32(15 * getHeightFactor()))
        smpDescTextBox.Location = New Point(bX, mapList.Y)
        smpDescTextBox.Size = New Size(cWidth, mapList.Height - 2 * bHeight - Convert.ToInt32(60 * getHeightFactor()))
        mapList.ItemHeight = Convert.ToInt32(25 * Game.getSizeFactor)
    End Sub

    Private Sub resizeWorldEditMenu()
        weHeader.Size = New Size(worldEditMenu.Width, Convert.ToInt32(worldEditMenu.Height / 4.0F))
        Dim bWidth As Integer = Convert.ToInt32(600 * getWidthFactor())
        Dim bHeight As Integer = Convert.ToInt32(50 * getHeightFactor())
        Dim bX As Integer = (mainMenu.Width - bWidth) \ 2
        weButton1.Size = New Size(bWidth, bHeight)
        weButton1.Location = New Point(bX, Convert.ToInt32(460 * getHeightFactor()))
        weButton2.Size = New Size(bWidth, bHeight)
        weButton2.Location = New Point(bX, Convert.ToInt32(600 * getHeightFactor()))
        weMapList.Location = New Point(bX, Convert.ToInt32(180 * getHeightFactor()))
        weMapList.Size = New Size(bWidth, Convert.ToInt32(260 * getWidthFactor()))
        weMapList.ItemHeight = Convert.ToInt32(25 * Game.getSizeFactor)
    End Sub

    Private Sub editWorld(name As String)
        Dim wec As New WorldEditControl(name)
        wec.parentMenu = worldEditMenu
        changeMenu(wec)
    End Sub

    Friend Function getEnvironmentPath(key As String) As String
        Return environmentPaths.Item(key.ToLower)
    End Function

    Friend Function loadBitmap(path As String) As Bitmap
        Dim GDI As New Drawing.Bitmap(path)
        Dim BMP As Bitmap = window.LoadBitmap(GDI)
        GDI.Dispose()
        Return BMP
    End Function

    Private Sub window_Resize(sender As Object, e As EventArgs) Handles window.Resize
        If mainMenu Is Nothing OrElse window.Minimized Then Exit Sub
        versionLabel.Font = FontManager.GetFont("versionFormat")
        omLabel1.Font = FontManager.GetFont("defaultLabelFormat")
        omLabel2.Font = FontManager.GetFont("defaultLabelFormat")
        omLabel3.Font = FontManager.GetFont("defaultLabelFormat")
        omHeader.Font = FontManager.GetFont("headerFormat")
        spmHeader.Font = FontManager.GetFont("headerFormat")
        weHeader.Font = FontManager.GetFont("headerFormat")
    End Sub

    Friend Sub changeMenu(m As Menu)
        If visibleMenu IsNot Nothing Then visibleMenu.Visible = False
        visibleMenu = m
        visibleMenu.Visible = True
        lastMenuChange = DateTime.Now
        window.ChangeCursor(0)
    End Sub

    Private Sub shutDown()
        Thread.Sleep(100)
        window.Close()
    End Sub

    Private Sub fullscreenButton_Click(sender As Object, e As EventArgs) Handles fullscreenButton.Click
        If (DateTime.Now - lastMenuChange).Ticks < 100000 Then Return
        window.setFullscreen(Not window.getFullscreen)
        If window.getFullscreen Then
            fullscreenButton.Text = guiINI.getValue("wm_Windowed")
        Else
            fullscreenButton.Text = guiINI.getValue("wm_Fullscreen")
        End If
    End Sub

    Private Sub mapList_SelectedIndexChanged(sender As Object, e As EventArgs) Handles mapList.SelectedIndexChanged
        If smpDescTextBox Is Nothing Then Exit Sub
        If mapList.SelectedIndex < 0 Then
            smpDescTextBox.Text = String.Empty
        Else
            smpDescTextBox.Text = worlds.Item(mapList.SelectedIndex).description
        End If
    End Sub

    Private Sub spmButton1_Click(sender As Object, e As EventArgs) Handles spmButton1.Click
        If mapList.SelectedIndex < 0 Then Exit Sub
        Dim wc As New WorldControl(worlds.Item(mapList.SelectedIndex))
        wc.parentMenu = singleplayerMenu
        startWorld(wc)
    End Sub
End Module

Friend Class Menu
    Inherits D2DPanel

    Friend parentMenu As Menu

    Friend Event show(sender As Object, e As EventArgs)
    Friend Event hide(sender As Object, e As EventArgs)

    Sub New()
        D2DParent = Game.window
        Visible = False
        setSize()
        AddHandler Game.window.Resize, Sub() setSize()
    End Sub

    Private Sub setSize()
        Dim rect As Rectangle = Game.window.getClientRect
        Location = rect.Location
        Size = rect.Size
    End Sub

    Friend Sub goUp()
        If parentMenu IsNot Nothing Then Game.changeMenu(parentMenu)
    End Sub

    Private Sub Menu_VisibleChanged(sender As Object, e As EventArgs) Handles Me.VisibleChanged
        If Visible Then
            RaiseEvent show(Me, EventArgs.Empty)
        Else
            RaiseEvent hide(Me, EventArgs.Empty)
        End If
    End Sub
End Class

Friend Class CriticalGameException
    Inherits Exception

    Sub New(message As String)
        MyBase.New(message)
    End Sub
End Class

Friend NotInheritable Class ColorManager
    Private Shared brushes As Dictionary(Of String, Brush)

    Friend Shared ReadOnly Property BrushRegistered(key As String) As Boolean
        Get
            Return brushes.ContainsKey(key)
        End Get
    End Property

    Shared Sub New()
        brushes = New Dictionary(Of String, Brush)
    End Sub

    Friend Shared Sub RegisterBrush(key As String, b As Brush)
        If BrushRegistered(key) Then Throw New BrushAlreadyRegisteredException(key)
        brushes.Add(key, b)
    End Sub

    Friend Shared Sub UnregisterBrush(key As String)
        brushes.Item(key).Dispose()
        brushes.Remove(key)
    End Sub

    Friend Shared Function GetBrush(key As String) As Brush
        Return brushes.Item(key)
    End Function

    Friend Shared Sub ClearAll()
        For Each item In brushes
            item.Value.Dispose()
        Next
        brushes.Clear()
        brushes = Nothing
    End Sub
End Class

Friend Class BrushAlreadyRegisteredException : Inherits Exception
    Private key_ As String

    Friend ReadOnly Property Key As String
        Get
            Return key_
        End Get
    End Property

    Friend Sub New(key As String)
        MyBase.New("Der angegebene Brush ist bereits registriert.")
        key_ = key
    End Sub
End Class

Friend NotInheritable Class FontManager
    Private Shared fonts As Dictionary(Of String, FontTemplate)

    Friend Shared ReadOnly Property FontRegistered(key As String) As Boolean
        Get
            Return fonts.ContainsKey(key)
        End Get
    End Property

    Shared Sub New()
        fonts = New Dictionary(Of String, FontTemplate)
    End Sub

    Friend Shared Sub RegisterFont(key As String, template As FontTemplate)
        If FontRegistered(key) Then Throw New FontAlreadyRegisteredException(key)
        fonts.Add(key, template)
    End Sub

    Friend Shared Sub UnrgisterFont(key As String)
        fonts.Item(key).Dispose()
        fonts.Remove(key)
    End Sub

    Friend Shared Function GetFont(key As String) As TextFormat
        Return fonts.Item(key).GetTextFormat(Game.getSizeFactor)
    End Function

    Friend Shared Sub ClearAll()
        For Each item In fonts
            item.Value.Dispose()
        Next
        fonts.Clear()
        fonts = Nothing
    End Sub
End Class

Friend Class FontAlreadyRegisteredException : Inherits Exception
    Private key_ As String

    Friend ReadOnly Property Key As String
        Get
            Return key_
        End Get
    End Property

    Friend Sub New(key As String)
        MyBase.New("Das angegebene FontTemplate ist bereits registriert.")
        key_ = key
    End Sub
End Class

Friend Class FontTemplate : Implements IDisposable
    Private font As TextFormat

    Private name As String
    Private size As Single
    Private weight As FontWeight
    Private style As FontStyle
    Private tAlign As TextAlignment
    Private pAlign As ParagraphAlignment

    Private factor As Single

    Friend Function GetTextFormat(sizeFactor As Single) As TextFormat
        If sizeFactor <> factor Then
            If font IsNot Nothing AndAlso Not font.Disposed Then font.Dispose()
            font = Game.window.getTextFactory.CreateTextFormat(name, weight, style, FontStretch.Normal, size * sizeFactor, Game.locale)
            font.TextAlignment = tAlign
            font.ParagraphAlignment = pAlign
            factor = sizeFactor
        End If
        Return font
    End Function

    Sub New(fontName As String, fontSize As Single, Optional fontStyle As FontStyle = DirectWrite.FontStyle.Normal, Optional fontWeight As FontWeight = DirectWrite.FontWeight.Normal, Optional textAlignment As TextAlignment = TextAlignment.Leading, Optional paragraphAlignment As ParagraphAlignment = ParagraphAlignment.Near)
        name = fontName
        size = fontSize
        weight = fontWeight
        style = fontStyle
        tAlign = textAlignment
        pAlign = paragraphAlignment
    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        If font IsNot Nothing Then font.Dispose()
    End Sub
End Class