﻿Option Strict On
Option Infer On

Imports D2DControlset
Imports SlimDX
Imports SlimDX.Direct2D
Imports System.IO
Imports System.Runtime.InteropServices

Friend Class GameWindow
    Inherits Direct2DWindow

    Private bg As BitmapBrush
    Private bgPos As Integer
    Private bgShine As LinearGradientBrush
    Private bgGlow As SolidColorBrush
    Private bgGlowOp As Integer
    Private bgGlowDir As Boolean
    Private shadow As LinearGradientBrush
    Private fullscreen As Boolean
    Private maximizedBeforeFullscreen As Boolean
    Private gameCursors As List(Of Cursor)
    Private cursorIndex As Integer

    <DllImport("user32.dll")>
    Private Shared Function LoadCursorFromFile(lpFileName As String) As IntPtr
    End Function

    Sub New()

        ' Dieser Aufruf ist für den Designer erforderlich.
        InitializeComponent()

        ' Fügen Sie Initialisierungen nach dem InitializeComponent()-Aufruf hinzu.
        fullscreen = False
        maximizedBeforeFullscreen = False
        cursorIndex = -1
        gameCursors = loadCursors(Game.getEnvironmentPath("cur"))
    End Sub

    Private Function loadCursors(directoryName As String) As List(Of Cursor)
        Dim dir As New DirectoryInfo(directoryName)
        If Not dir.Exists Then Return Nothing
        Dim l As New List(Of Cursor)
        For Each fi In dir.GetFiles
            Dim cur = loadCursor(fi)
            If cur IsNot Nothing Then l.Add(cur)
        Next
        Return l
    End Function

    Private Function loadCursor(fi As FileInfo) As Cursor
        If Not (fi.Exists AndAlso (fi.Extension = ".cur" OrElse fi.Extension = ".ani")) Then Return Nothing
        Dim hWndCursor = LoadCursorFromFile(fi.FullName)
        If hWndCursor = IntPtr.Zero Then Return Nothing
        Dim cur As New Cursor(hWndCursor)
        Return cur
    End Function

    Friend Function getTarget() As Direct2D.RenderTarget
        Return Target
    End Function

    Friend Function getFactory() As Direct2D.Factory
        Return Factory
    End Function

    Friend Function getTextFactory() As DirectWrite.Factory
        Return TextFactory
    End Function

    Friend Function getClientRect() As Rectangle
        Dim rect As New Rectangle
        If fullscreen Then
            rect.Location = New Point(0, 0)
            rect.Size = Size
        Else
            If Maximized Then
                rect.Location = New Point(0, 30)
                rect.Size = New Size(Width, Height - 30)
            Else
                rect.Location = New Point(8, 30)
                rect.Size = New Size(Width - 16, Height - 38)
            End If
        End If
        Return rect
    End Function

    Friend Sub setFullscreen(ByVal val As Boolean)
        If val = fullscreen Then Exit Sub
        If val Then
            Me.ShowWindowBorder = False
            Me.Hide()
            If Maximized Then
                maximizedBeforeFullscreen = True
                Me.MaximizeOrRestore()
            Else
                maximizedBeforeFullscreen = False
            End If
            Me.MaximizeOrRestore()
            Me.Show()
        Else
            Me.Hide()
            Me.ShowWindowBorder = True
            Me.MaximizeOrRestore()
            If maximizedBeforeFullscreen Then Me.MaximizeOrRestore()
            Me.Show()
        End If
        fullscreen = val
    End Sub

    Friend Function getFullscreen() As Boolean
        Return fullscreen
    End Function

    Protected Overrides Sub InitD2D()
        MyBase.InitD2D()
        loadBackground()
    End Sub

    Protected Overrides Sub UnloadD2D()
        unloadBackground()
        MyBase.UnloadD2D()
    End Sub

    Private Sub loadBackground()
        Dim GDIbg As Drawing.Bitmap = New Drawing.Bitmap(Path.Combine(Game.getEnvironmentPath("img"), "bg.png"))
        Dim D2Dbg As Bitmap = LoadBitmap(GDIbg)
        Dim props As New BitmapBrushProperties
        props.InterpolationMode = InterpolationMode.Linear
        props.VerticalExtendMode = ExtendMode.Wrap
        props.HorizontalExtendMode = ExtendMode.Wrap
        bg = New BitmapBrush(Target, D2Dbg, props)
        D2Dbg.Dispose()
        GDIbg.Dispose()
        Dim shineColor1 As New Color4(Color.FromArgb(255, 208, 128, 229))
        Dim shineColor2 As New Color4(0, 0, 0, 0)
        Dim shineColor3 As New Color4(Color.FromArgb(64, 208, 128, 229))
        Dim s1 As New GradientStop
        s1.Color = shineColor2
        s1.Position = 0
        Dim s2 As New GradientStop
        s2.Color = shineColor3
        s2.Position = 0.35F
        Dim s3 As New GradientStop
        s3.Color = shineColor1
        s3.Position = 0.5F
        Dim s4 As New GradientStop
        s4.Color = shineColor3
        s4.Position = 0.65F
        Dim s5 As New GradientStop
        s5.Color = shineColor2
        s5.Position = 1
        Dim stops As New GradientStopCollection(Target, {s1, s2, s3, s4, s5})
        Dim shineProps As New LinearGradientBrushProperties
        bgPos = Me.Height + 200
        shineProps.StartPoint = New PointF(0, bgPos - 200)
        shineProps.EndPoint = New PointF(0, bgPos)
        bgShine = New LinearGradientBrush(Target, stops, shineProps)
        stops.Dispose()
        bgGlow = New SolidColorBrush(Target, shineColor1)
        bgGlowOp = 280
        bgGlow.Opacity = bgGlowOp / 1600.0F
        Dim sStop1 As New GradientStop
        sStop1.Color = New Color4(1, 0, 0, 0)
        sStop1.Position = 0
        Dim sStop2 As New GradientStop
        sStop2.Color = New Color4(0, 0, 0, 0)
        sStop2.Position = 1
        Dim sStops As New GradientStopCollection(Target, {sStop1, sStop2})
        Dim sProps As New LinearGradientBrushProperties
        sProps.StartPoint = New Point(0, 30)
        sProps.EndPoint = New Point(0, 60)
        shadow = New LinearGradientBrush(Target, sStops, sProps)
        sStops.Dispose()
    End Sub

    Private Sub unloadBackground()
        bgGlow.Dispose()
        bgShine.Dispose()
        bg.Dispose()
        shadow.Dispose()
    End Sub

    Private Sub drawBackground()
        If bgShine Is Nothing Then Exit Sub
        bgPos -= 1
        If bgPos < 0 Then bgPos = Me.Height + 200
        bgShine.StartPoint = New PointF(0, bgPos - 200)
        bgShine.EndPoint = New PointF(0, bgPos)
        If bgGlowDir Then
            bgGlowOp += 1
            If bgGlowOp >= 280 Then bgGlowDir = False
        Else
            bgGlowOp -= 1
            If bgGlowOp <= 140 Then bgGlowDir = True
        End If
        bgGlow.Opacity = bgGlowOp / 1600.0F
        Target.FillRectangle(bgGlow, Me.ClientRectangle)
        Target.FillRectangle(bgShine, New Rectangle(0, bgPos - 200, Me.Width, 200))
        Target.FillRectangle(bg, Me.ClientRectangle)
        If ShowWindowBorder Then Target.FillRectangle(shadow, New Rectangle(0, 30, Me.Width, 30))
    End Sub

    Private Sub GameWindow_D2DDraw(sender As Object, e As D2DDrawEventArgs) Handles Me.D2DDraw
        drawBackground()
    End Sub

    Friend Shadows Function measureString(s As String, f As DirectWrite.TextFormat) As SizeF
        Return MyBase.MeasureString(s, f)
    End Function

    Protected Overrides Sub RequestCursorChange(c As Cursor)
        Select c
            Case Cursors.Default
                ChangeCursor(0)
            Case Else
                MyBase.RequestCursorChange(c)
        End Select
    End Sub

    Friend Sub ChangeCursor(index As Integer)
        If gameCursors(index) = Cursor Then Exit Sub
        If index >= gameCursors.Count OrElse index < 0 Then
            ChangeCursor(0)
        Else
            Cursor = gameCursors(index)
            cursorIndex = index
        End If
    End Sub
End Class