﻿Option Strict On
Option Infer On

Imports SlimDX
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports D2DControlset
Imports Helpers.Direct2D
Imports Helpers.Mathematics

Friend Class InfoMenu
    Inherits GameControl

    Private upButton As IngameButton

    Friend Event upgradeRequested(sender As Object, e As EventArgs)
    Friend Event upgradeCostShow(sender As Object, e As EventArgs)
    Friend Event upgradeCostHide(sender As Object, e As EventArgs)

    Friend Property towerName As String

    Friend Property towerLevel As Integer

    Friend Property towerRadius As Integer

    Friend Property towerDamage As Integer

    Protected Overrides Sub drawContent(e As D2DDrawEventArgs, inRect As RectangleF)
        Dim f = Game.getSizeFactor
        Dim absPos = AbsoluteLocation
        Dim nameRect As New RectangleF(absPos.X + 10 * f, absPos.Y + 6 * f, Width - 20 * f, 24 * f)
        Dim nameFormat = e.TextFactory.CreateTextFormat("Segoe UI", FontWeight.Medium, FontStyle.Normal, FontStretch.Normal, 24 * f, Game.locale)
        nameFormat.TextAlignment = TextAlignment.Center
        e.Target.DrawText(towerName, nameFormat, nameRect, ColorManager.GetBrush("defaultWhite"))
        nameFormat.Dispose()
        Dim infoRect As New RectangleF(nameRect.X, nameRect.Y + nameRect.Height + 10 * f, nameRect.Width, 30 * f)
        Dim infoFormat = e.TextFactory.CreateTextFormat("Segoe UI", FontWeight.Normal, FontStyle.Normal, FontStretch.Normal, 16 * f, Game.locale)
        infoFormat.ParagraphAlignment = ParagraphAlignment.Center
        infoFormat.TextAlignment = TextAlignment.Center
        e.Target.DrawText(String.Join("   ", "Level: " & towerLevel, "Radius: " & towerRadius, "Schaden: " & towerDamage), infoFormat, infoRect, ColorManager.GetBrush("defaultWhite"))
        infoFormat.Dispose()
        Dim bmpHeight = Height - 20 * f
        Dim bmpWidth = bmpHeight / handlingBMP.PixelSize.Height * handlingBMP.PixelSize.Width
        e.Target.DrawBitmap(Game.handlingBMP, New RectangleF(absPos.X + 20 * f, absPos.Y + 10 * f, bmpWidth, bmpHeight))
    End Sub

    Sub New()
        upButton = New IngameButton
        upButton.D2DParent = Me
        upButton.Text = "Upgrade [U]"
        resizeButton()
        AddHandler upButton.Click, Sub()
                                       RaiseEvent upgradeRequested(Me, EventArgs.Empty)
                                       RaiseEvent upgradeCostShow(Me, EventArgs.Empty)
                                   End Sub
        AddHandler upButton.MouseEnter, Sub() RaiseEvent upgradeCostShow(Me, EventArgs.Empty)
        AddHandler upButton.MouseLeave, Sub() RaiseEvent upgradeCostHide(Me, EventArgs.Empty)
        AddHandler Game.window.KeyDown, Sub(sender, e) If e.KeyCode = Keys.U Then RaiseEvent upgradeRequested(Me, EventArgs.Empty)
        AddHandler Me.Resize, Sub() resizeButton()
    End Sub

    Private Sub resizeButton()
        Dim f = Game.getSizeFactor
        upButton.Width = Convert.ToInt32(140 * f)
        upButton.Height = Convert.ToInt32(35 * f)
        upButton.X = Width - upButton.Width - Convert.ToInt32(20 * f)
        upButton.Y = (Height - upButton.Height) \ 2
    End Sub
End Class

Friend Class InfoPicture
    Inherits GameControl

    Friend Property image As Bitmap

    Protected Overrides Sub drawContent(e As D2DDrawEventArgs, inRect As RectangleF)
        If image IsNot Nothing Then e.Target.DrawBitmap(image, inRect, If(m_Entered, 0.3F, 1.0F))
    End Sub
End Class

Friend Class InfoSidebar
    Inherits GameControl

    Private m_life As Integer
    Private lifeInfo As LifeInfo

    Friend Property money As Integer

    Friend Property toPay As Integer

    Friend Property waveInfo As String

    Friend Property life As Integer
        Get
            Return m_life
        End Get
        Set(value As Integer)
            Dim dif = m_life - value
            For i = 1 To dif
                lifeInfo.removeLife()
            Next
            m_life = value
        End Set
    End Property

    Sub New()
        lifeInfo = New LifeInfo(Me)
        resizeLifeInfo()
    End Sub

    Protected Overrides Sub drawContent(e As D2DDrawEventArgs, inRect As RectangleF)
        Dim f = Game.getSizeFactor
        Dim availableHeight = Height - 2 * borderWidth
        Dim itemHeight = availableHeight / 3.0F
        Dim textHeight = itemHeight * 0.7F
        Dim addPoint As New PointF(AbsoluteLocation.X + borderWidth, AbsoluteLocation.Y + borderWidth)
        Dim lineWidth = Width - 2 * borderWidth
        e.Target.DrawLine(ColorManager.GetBrush("defaultWhite"), New PointF(addPoint.X, addPoint.Y + itemHeight), New PointF(addPoint.X + lineWidth, addPoint.Y + itemHeight), 2 * f)
        e.Target.DrawLine(ColorManager.GetBrush("defaultWhite"), New PointF(addPoint.X, addPoint.Y + 2 * itemHeight), New PointF(addPoint.X + lineWidth, addPoint.Y + 2 * itemHeight), 2 * f)
        Dim format = e.TextFactory.CreateTextFormat("Segoe UI", FontWeight.Normal, FontStyle.Normal, FontStretch.Normal, textHeight, Game.locale)
        format.ParagraphAlignment = ParagraphAlignment.Center
        Dim textX = 8 * f
        Dim textWidth = lineWidth - 2 * textX
        Dim rect1 As New RectangleF(addPoint.X + textX, addPoint.Y, textWidth, itemHeight)
        e.Target.DrawText(waveInfo, format, rect1, ColorManager.GetBrush("defaultWhite"))
        Dim rect2 As New RectangleF(addPoint.X + textX, addPoint.Y + itemHeight, textWidth, itemHeight)
        e.Target.DrawText(money & "$", format, rect2, ColorManager.GetBrush("defaultWhite"))
        If toPay > 0 Then
            format.TextAlignment = TextAlignment.Trailing
            Dim red As New SolidColorBrush(e.Target, New Color4(1, 0, 0))
            Dim green As New SolidColorBrush(e.Target, New Color4(0, 1, 0))
            Dim usedBrush As SolidColorBrush = green
            If toPay > money Then usedBrush = red
            e.Target.DrawText(toPay & "$", format, rect2, usedBrush)
            red.Dispose()
            green.Dispose()
        End If
        format.Dispose()
    End Sub

    Private Sub InfoSidebar_Resize(sender As Object, e As EventArgs) Handles Me.Resize
        resizeLifeInfo()
    End Sub

    Private Sub resizeLifeInfo()
        Dim availableHeight = Height - 2 * borderWidth
        lifeInfo.Location = New Point(Convert.ToInt32(borderWidth), Convert.ToInt32(availableHeight * 2 / 3 + borderWidth))
        lifeInfo.Size = New Size(Convert.ToInt32(Width - 2 * borderWidth), Convert.ToInt32(availableHeight / 3))
    End Sub
End Class

Friend Class LifeInfo
    Inherits D2DControl

    Private stars As List(Of LifeStar)
    Private lastIndex As Integer

    Protected Overrides Sub OnDraw(e As D2DDrawEventArgs)

    End Sub

    Sub New(parent As IDirect2DControl)
        MyBase.New(parent)
        stars = New List(Of LifeStar)
        For i = 0 To 9
            Dim star As New LifeStar(Me)
        Next
        lastIndex = 9
    End Sub

    Friend Sub removeLife()
        If lastIndex >= 0 Then
            stars.Item(lastIndex).lifeLost()
            lastIndex -= 1
        End If
    End Sub

    Private Class LifeStar
        Private info As LifeInfo
        Private index As Integer
        Private rotation As Integer
        Private removing As Boolean
        Private size As Single
        Private location As PointF
        Private removeStep As Integer

        Sub New(info As LifeInfo)
            Me.info = info
            index = info.stars.Count
            removeStep = 40
            removing = False
            info.stars.Add(Me)
            AddHandler info.D2DDraw, AddressOf doStep
            AddHandler info.Resize, AddressOf resize
        End Sub

        Private Sub doStep(sender As Object, e As D2DDrawEventArgs)
            If removing Then
                removeStep -= 1
                Dim starDistance = info.Width / 11.0F
                size = starDistance * 0.8F * (removeStep / 40.0F)
            End If
            If removeStep <= 0 Then
                remove()
            Else
                rotation += 1
                rotation = rotation Mod 360
                draw(e)
            End If
        End Sub

        Private Sub draw(e As D2DDrawEventArgs)
            Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
            h.FillStar(New PointF(location.X + info.AbsoluteLocation.X, location.Y + info.AbsoluteLocation.Y), size, -rotation, ColorManager.GetBrush("defaultWhite"), 5)
        End Sub

        Private Sub resize(sender As Object, e As EventArgs)
            Dim starDistance = info.Width / 11.0F
            location = New PointF((index + 1) * starDistance, info.Height / 2.0F)
            size = starDistance * 0.45F
        End Sub

        Private Sub remove()
            RemoveHandler info.D2DDraw, AddressOf doStep
            RemoveHandler info.Resize, AddressOf resize
            info.stars.Remove(Me)
            info = Nothing
        End Sub

        Friend Sub lifeLost()
            removing = True
        End Sub
    End Class
End Class