﻿Option Strict On
Option Infer On

Imports D2DControlset
Imports SlimDX
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports System.IO

Friend Class TDButton
    Inherits D2DControl

    Private sound1 As SoundEffect
    Private sound2 As SoundEffect

    Protected Overrides Sub OnDraw(e As D2DDrawEventArgs)
        If Not String.IsNullOrWhiteSpace(Text) Then
            Dim layout As TextLayout = e.TextFactory.CreateTextLayout(Text, FontManager.GetFont("defaultFormat"), Width, Height)
            layout.ParagraphAlignment = ParagraphAlignment.Center
            layout.TextAlignment = TextAlignment.Center
            e.Target.DrawTextLayout(AbsoluteLocation, layout, ColorManager.GetBrush("defaultWhite"))
            layout.Dispose()
        End If
        If Not m_Entered Then Exit Sub
        Dim gPath As New PathGeometry(e.Factory)
        Dim sink = gPath.Open
        Dim pPoints(8) As PointF
        Dim b1 = Me.Height * (2.0F / 2.5F)
        Dim b2 = Me.Height * 1.5F
        Dim b3 = 2.5F * Game.getSizeFactor
        Dim x1 = b3 + Me.AbsoluteLocation.X
        Dim x2 = b1 + Me.AbsoluteLocation.X
        Dim x3 = b2 + Me.AbsoluteLocation.X
        Dim x4 = Me.Width - b2 + Me.AbsoluteLocation.X
        Dim x5 = Me.Width - b1 + Me.AbsoluteLocation.X
        Dim x6 = Me.Width - b3 + Me.AbsoluteLocation.X
        Dim y1 = b3 + Me.AbsoluteLocation.Y
        Dim y2 = Me.Height / 4.0F + Me.AbsoluteLocation.Y
        Dim y3 = Me.Height / 2.0F + Me.AbsoluteLocation.Y
        Dim y4 = Me.Height / 4.0F * 3.0F + Me.AbsoluteLocation.Y
        Dim y5 = Me.Height - b3 + Me.AbsoluteLocation.Y
        sink.BeginFigure(New PointF(x3, y1), FigureBegin.Hollow)
        pPoints(0) = New PointF(x4, y1)
        pPoints(1) = New PointF(x5, y4)
        pPoints(2) = New PointF(x6, y3)
        pPoints(3) = New PointF(x5, y2)
        pPoints(4) = New PointF(x4, y5)
        pPoints(5) = New PointF(x3, y5)
        pPoints(6) = New PointF(x2, y2)
        pPoints(7) = New PointF(x1, y3)
        pPoints(8) = New PointF(x2, y4)
        sink.AddLines(pPoints)
        sink.EndFigure(FigureEnd.Closed)
        sink.Close()
        e.Target.DrawGeometry(gPath, ColorManager.GetBrush("defaultWhite"), 2.5F * Game.getSizeFactor)
        sink.Dispose()
        gPath.Dispose()
    End Sub

    Sub New()
        sound1 = New SoundEffect("bt01.wma")
        sound2 = New SoundEffect("bt02.wma")
    End Sub

    Private Sub TDButton_Click(sender As Object, e As EventArgs) Handles Me.Click
        sound1.play()
    End Sub

    Private Sub TDButton_MouseEnter(sender As Object, e As EventArgs) Handles Me.MouseEnter
        sound2.play()
        Game.window.ChangeCursor(1)
    End Sub

    Private Sub TDButton_MouseLeave(sender As Object, e As EventArgs) Handles Me.MouseLeave
        Game.window.ChangeCursor(0)
    End Sub
End Class
