﻿Imports SlimDX
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports D2DControlset
Imports System.ComponentModel
Imports System.Reflection

Friend Class TDListBox : Inherits D2DControl
    Private scrollBar As TDScrollBar
    Private m_items As ListBoxItemCollection
    Private m_SelectedIndex As Integer
    Private m_itemHeight As Integer

    Friend Event SelectedIndexChanged(sender As Object, e As EventArgs)
    Private Event OnDrawItems(e As D2DDrawEventArgs)

    Friend ReadOnly Property Items As ListBoxItemCollection
        Get
            Return m_items
        End Get
    End Property

    Friend Property SelectedIndex As Integer
        Get
            Return m_SelectedIndex
        End Get
        Set(value As Integer)
            If value = m_SelectedIndex Then Return
            If value < 0 OrElse value >= m_items.Count Then
                m_SelectedIndex = -1
            Else
                m_SelectedIndex = value
            End If
            RaiseEvent SelectedIndexChanged(Me, EventArgs.Empty)
        End Set
    End Property

    Friend Property SelectedItem As ListBoxItem
        Get
            If SelectedIndex < 0 OrElse SelectedIndex >= m_items.Count Then Return Nothing
            Return m_items(SelectedIndex)
        End Get
        Set(value As ListBoxItem)
            If value Is SelectedItem Then Return
            If value Is Nothing Then SelectedIndex = -1
            Dim i = m_items.IndexOf(value)
            SelectedIndex = i
        End Set
    End Property

    Friend Property ItemHeight As Integer
        Get
            Return m_itemHeight
        End Get
        Set(value As Integer)
            m_itemHeight = Math.Max(value, 1)
            Dim percentage = scrollBar.Value / scrollBar.MaxValue
            scrollBar.MaxValue = Math.Max(m_items.Count * ItemHeight - GetItemRect.Height, 0)
            If scrollBar.MaxValue > 0 Then scrollBar.Value = Convert.ToInt32(scrollBar.MaxValue * percentage)
        End Set
    End Property

    Protected Overrides Sub OnDraw(e As D2DDrawEventArgs)
        DrawBorder(e)
        Dim rect = GetItemRect()
        Dim absLoc = AbsoluteLocation
        rect.X += absLoc.X
        rect.Y += absLoc.Y
        e.Target.PushAxisAlignedClip(rect, AntialiasMode.PerPrimitive)
        RaiseEvent OnDrawItems(e)
        e.Target.PopAxisAlignedClip()
    End Sub

    Private Sub DrawBorder(e As D2DDrawEventArgs)
        Dim lineWidth = 2.0F * Game.getSizeFactor
        Dim cornerSize = scrollBar.Width / 2.0F
        Dim absLoc = AbsoluteLocation
        Dim x1 = lineWidth / 2.0F + absLoc.X
        Dim x2 = x1 + cornerSize
        Dim x4 = Width - scrollBar.Width - lineWidth + absLoc.X
        Dim x3 = x4 - cornerSize
        Dim y1 = lineWidth / 2.0F + absLoc.Y
        Dim y2 = y1 + cornerSize
        Dim y4 = Height - lineWidth / 2.0F + absLoc.Y
        Dim y3 = y4 - cornerSize
        Dim points(7) As PointF
        points(0) = New PointF(x1, y2)
        points(1) = New PointF(x2, y1)
        points(2) = New PointF(x3, y1)
        points(3) = New PointF(x4, y2)
        points(4) = New PointF(x4, y3)
        points(5) = New PointF(x3, y4)
        points(6) = New PointF(x2, y4)
        points(7) = New PointF(x1, y3)
        Dim pg As New PathGeometry(e.Factory)
        Dim sink = pg.Open
        sink.BeginFigure(points.First, FigureBegin.Hollow)
        sink.AddLines(points.Skip(1).ToArray)
        sink.EndFigure(FigureEnd.Closed)
        sink.Close()
        e.Target.DrawGeometry(pg, ColorManager.GetBrush("defaultWhite"), lineWidth)
        sink.Dispose()
        pg.Dispose()
    End Sub

    Private Function GetItemRect() As Rectangle
        Dim border = Convert.ToInt32(4.0F * Game.getSizeFactor + scrollBar.Width / 2.0F)
        Return New Rectangle(border, border, Width - 2 * border - scrollBar.Width - Convert.ToInt32(2.0F * Game.getSizeFactor), Height - 2 * border)
    End Function

    Sub New()
        scrollBar = New TDScrollBar
        scrollBar.D2DParent = Me
        ResizeScrollBar()
        AddHandler Me.Resize, Sub() ResizeScrollBar()
        m_itemHeight = 20
        m_items = New ListBoxItemCollection
        m_SelectedIndex = -1
        AddHandler m_items.ListChanged, _
            Sub(sender, e)
                Select Case e.ListChangedType
                    Case ListChangedType.ItemAdded
                        m_items(e.NewIndex).ListBox = Me
                        If e.NewIndex <= m_SelectedIndex Then SelectedIndex += 1
                    Case ListChangedType.ItemMoved
                        If e.OldIndex > m_SelectedIndex AndAlso e.NewIndex <= m_SelectedIndex Then SelectedIndex += 1
                        If e.OldIndex < m_SelectedIndex AndAlso e.NewIndex >= m_SelectedIndex Then SelectedIndex -= 1
                    Case ListChangedType.ItemDeleted
                        If e.NewIndex = SelectedIndex Then SelectedIndex = -1
                        If e.NewIndex < SelectedIndex Then SelectedIndex -= 1
                    Case ListChangedType.Reset
                        SelectedIndex = -1
                End Select
                scrollBar.MaxValue = Math.Max(m_items.Count * ItemHeight - GetItemRect.Height, 0)
            End Sub
        AddHandler m_items.ItemRemoving, Sub(index) m_items(index).ListBox = Nothing
    End Sub

    Private Sub ResizeScrollBar()
        Dim scrollBarWidth = Convert.ToInt32(25.0F * Game.getSizeFactor)
        scrollBar.Size = New Size(scrollBarWidth, Height)
        scrollBar.Location = New Point(Width - scrollBarWidth, 0)
        If m_items IsNot Nothing Then scrollBar.MaxValue = Math.Max(m_items.Count * ItemHeight - GetItemRect.Height, 0)
    End Sub

    Friend MustInherit Class ListBoxItem
        Protected m_listBox As TDListBox
        Private selected As Boolean

        Friend Property Text As String

        Friend Property ListBox As TDListBox
            Get
                Return m_listBox
            End Get
            Set(value As TDListBox)
                If value Is m_listBox Then Return
                If m_listBox IsNot Nothing Then
                    RemoveHandler m_listBox.OnDrawItems, AddressOf ListBox_OnDrawItems
                    RemoveHandler m_listBox.SelectedIndexChanged, AddressOf ListBox_SelectedIndexChanged
                    m_listBox = Nothing
                End If
                If value IsNot Nothing Then
                    m_listBox = value
                    AddHandler m_listBox.OnDrawItems, AddressOf ListBox_OnDrawItems
                    AddHandler m_listBox.SelectedIndexChanged, AddressOf ListBox_SelectedIndexChanged
                End If
            End Set
        End Property

        Friend Property IsSelected As Boolean
            Get
                Return selected
            End Get
            Private Set(value As Boolean)
                selected = value
            End Set
        End Property

        Friend ReadOnly Property IsVisible As Boolean
            Get
                If m_listBox Is Nothing Then Return False
                Dim index = m_listBox.m_items.IndexOf(Me)
                Return (index + 1) * m_listBox.m_itemHeight > m_listBox.scrollBar.Value AndAlso index * m_listBox.m_itemHeight < m_listBox.scrollBar.Value + m_listBox.GetItemRect.Height
            End Get
        End Property

        Private Sub ListBox_OnDrawItems(e As D2DDrawEventArgs)
            If IsVisible Then
                Dim location = m_listBox.AbsoluteLocation
                Dim rect = m_listBox.GetItemRect
                Dim listOffset = rect.Location
                location.X += listOffset.X
                location.Y += listOffset.Y
                location.Y -= m_listBox.scrollBar.Value
                location.Y += m_listBox.m_items.IndexOf(Me) * m_listBox.ItemHeight
                DrawItem(New Rectangle(location, New Size(rect.Width, m_listBox.ItemHeight)), e)
            End If
        End Sub

        Private Sub ListBox_SelectedIndexChanged(sender As Object, e As EventArgs)
            IsSelected = m_listBox.SelectedItem Is Me
        End Sub

        Protected MustOverride Sub DrawItem(rect As Rectangle, e As D2DDrawEventArgs)

    End Class

    Friend Class ListBoxItemCollection : Inherits BindingList(Of ListBoxItem)
        Public Event ItemRemoving(index As Integer)

        Protected Overrides Sub RemoveItem(index As Integer)
            RaiseEvent ItemRemoving(index)
            MyBase.RemoveItem(index)
        End Sub
    End Class

    Private Sub TDListBox_MouseClick(sender As Object, e As MouseEventArgs) Handles Me.MouseClick
        If (DateTime.Now - Game.lastMenuChange).Ticks < 100000 Then Exit Sub
        Dim itemRect = GetItemRect()
        If itemRect.Contains(e.Location) Then
            Dim relativePos = e.Y + scrollBar.Value - itemRect.Y
            Dim index = relativePos \ m_itemHeight
            If index >= m_items.Count Then
                SelectedIndex = -1
            Else
                SelectedIndex = index
            End If
        Else
            SelectedIndex = -1
        End If
    End Sub
End Class

Friend Class TDTextOnlyListBoxItem : Inherits TDListBox.ListBoxItem

    Protected Overrides Sub DrawItem(rect As Rectangle, e As D2DDrawEventArgs)
        If Not String.IsNullOrWhiteSpace(Text) Then
            Dim formatName = "ListBoxItemFormat"
            If Not FontManager.FontRegistered(formatName) Then FontManager.RegisterFont(formatName, New FontTemplate("Polentical Neon", 16, ParagraphAlignment:=ParagraphAlignment.Center))
            e.Target.DrawText(Text, FontManager.GetFont(formatName), rect, ColorManager.GetBrush("defaultWhite"))
            If IsSelected Then e.Target.DrawRectangle(ColorManager.GetBrush("defaultWhite"), New RectangleF(rect.X + 0.5F, rect.Y + 0.5F, rect.Width - 1, rect.Height - 1))
        End If
    End Sub
End Class

Friend Class TDImageListBoxItem : Inherits TDListBox.ListBoxItem

    Friend Property Image As Bitmap

    Protected Overrides Sub DrawItem(rect As Rectangle, e As D2DDrawEventArgs)
        If Not String.IsNullOrWhiteSpace(Text) Then
            If Image IsNot Nothing Then
                Dim border = Convert.ToInt32(rect.Height * 0.1F)
                e.Target.DrawBitmap(Image, New Rectangle(rect.X + border, rect.Y + border, rect.Height - 2 * border, rect.Height - 2 * border))
            End If
            Dim formatName = "ListBoxItemFormat"
            If Not FontManager.FontRegistered(formatName) Then FontManager.RegisterFont(formatName, New FontTemplate("Polentical Neon", 16, ParagraphAlignment:=ParagraphAlignment.Center))
            Dim xOffset = rect.Height + Convert.ToInt32(10.0F * Game.getSizeFactor)
            Dim textRect As New Rectangle(rect.X + xOffset, rect.Y, rect.Width - xOffset, rect.Height)
            e.Target.DrawText(Text, FontManager.GetFont(formatName), textRect, ColorManager.GetBrush("defaultWhite"))
            If IsSelected Then e.Target.DrawRectangle(ColorManager.GetBrush("defaultWhite"), New RectangleF(textRect.X - 0.5F, textRect.Y + 0.5F, textRect.Width, textRect.Height - 1))
        End If
    End Sub
End Class